#include <Python.h>
#include "Basic_Types.h"
#include "MSP430mspgcc.h"

/* _parjtag extension.
part of http://mspgcc.sf.net
chris <cliechti@gmx.net>
*/

//JTAG functions
#define  __doc__connect "connect()\n"\
"Enable JTAG and connect to target. This stops it.\n"\
"This function must be called before using any other JTAG function,\n"\
"or the other functions will yield unpredictable data."
static PyObject*
py_connect(PyObject *self, PyObject *args)
{
    char *port;
    STATUS_T status;
    LONG version;
#ifdef WIN32
    port = "1";
#else
    port = "/dev/parport0";
#endif
    if (!PyArg_ParseTuple(args, "|s", &port))
        return 0;

    status = MSP430_Initialize(port, &version);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not initialize the library (port: %s)", port);
        return NULL;
    }

    MSP430_VCC(3);
    
    status = MSP430_Open();
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Can't open interface");
        return NULL;
    }

    status = MSP430_Configure(VERIFICATION_MODE, TRUE);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not configure the library");
        return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

#define  __doc__release "release()\n"\
"Release the target, disable JTAG lines.\n"\
"Subsequent access to the JTAG yields wrong data, until\n"\
"connect() is called again.\n"\
"The execution is started wherever the PC stays. Don't use this\n"\
"function after Flash operations or memverify. The PC was modified\n"\
"and points to an unpredicatble location. Use reset() before calling\n"\
"this function."
static PyObject*
py_release(PyObject *self, PyObject *args)
{
    STATUS_T status;
    if (!PyArg_ParseTuple(args, ""))
        return 0;
    status = MSP430_Close(TRUE);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not close the library");
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

#define  __doc__reset "reset(execute=0, release=0)\n"\
"Reset the device, optionaly start execution and/or release JTAG."
static PyObject*
py_reset(PyObject *self, PyObject *args)
{
    STATUS_T status;
    unsigned int release = 0;
    unsigned int execute = 0;
    unsigned int resets = ALL_RESETS;
    if (!PyArg_ParseTuple(args, "|iii", &execute, &release, &resets))
        return 0;
    status = MSP430_Reset(resets, execute, release);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not reset target (no connection?)");
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

#define  __doc__memread "memread(address, size)\n"\
"Read 'size' bytes starting at the specified address.\n"\
"The return value is a string with the (binary) data.\n"\
"It is possible to read peripherals, RAM as well as Flash.\n"
static PyObject*
py_memread(PyObject *self, PyObject *args)
{
    STATUS_T status;
    unsigned int address = 0;
    unsigned int size = 0;
    char * buffer;
    PyObject *res;
    
    if (!PyArg_ParseTuple(args, "ii", &address, &size))
        return 0;

    buffer = malloc(size+2);    //just to be sure + 2 (shouldn't be needed though)
    if (buffer==NULL) {
        PyErr_Format(PyExc_MemoryError, "no memory for temp buffer");
        return NULL;
    }
    
    status = MSP430_Memory(address, buffer, size, READ);
    if (status == STATUS_OK) {
        res = Py_BuildValue("s#", buffer, size);
    } else {
        PyErr_Format(PyExc_IOError, "Could not read target memory");
        res = NULL;
    }
    free(buffer);
    return res;
}

#define  __doc__memwrite "memwrite(address, mem)\n"\
"'mem' has to be a string, containing the data to write.\n"\
"It is possible to write peripherals, RAM as well as Flash.\n"\
"Flash must be erased before writing it with memerase()."
static PyObject*
py_memwrite(PyObject *self, PyObject *args)
{
    STATUS_T status;
    unsigned int address = 0;
    unsigned int size = 0;
    char * buffer = NULL;
    
    if (!PyArg_ParseTuple(args, "is#", &address, &buffer, &size))
        return 0;
    
    status = MSP430_Memory(address, buffer, size, WRITE);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not write target memory");
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

#define  __doc__memverify "memverify(address, mem)\n"\
"'mem' has to be a string of even length.\n"\
"Verify device memory aginst the supplied data using PSA analysis."
static PyObject*
py_memverify(PyObject *self, PyObject *args)
{
    STATUS_T status;
    unsigned int address = 0;
    unsigned int size = 0;
    PyObject *res;
    char * buffer = NULL;
    
    if (!PyArg_ParseTuple(args, "is#", &address, &buffer, &size))
        return 0;
    if (size&1) {
        PyErr_Format(PyExc_ValueError, "Buffer must have an even length");
        return NULL;
    }
    status = MSP430_VerifyMem(address, size, buffer);
    if (status == STATUS_OK) {
        res = Py_BuildValue("i", 1);            //equal/ok
    } else {
        res = Py_BuildValue("i", 0);            //not equal
    }
    return res;
}

#define  __doc__memerase "memerase(type=ERASE_ALL, address=0xfffe, length=2)\n"\
"Erase the Flash.\n\n"\
"Valid modes are:\n"\
"    ERASE_SGMT = 0\n"\
"    ERASE_MAIN = 1\n"\
"    ERASE_ALL  = 2\n\n"\
"The default address and length is fine for mass and main erase.\n"\
"To erase a single segment ERASE_SGMT and an address within that\n"\
"segement must be specified. The length can be choosen larger than\n"\
"one segment to erase a consecutive block of segments.\n"\
"The erased segments are checked for erasure using PSA analysis."
static PyObject*
py_memerase(PyObject *self, PyObject *args)
{
    STATUS_T status;
    unsigned int type = ERASE_ALL;
    unsigned int address = 0xfffe;
    unsigned int length = 2;
    if (!PyArg_ParseTuple(args, "|iii", &type, &address, &length))
        return 0;
    status = MSP430_Erase(type, address, length);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not erase the Flash");
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

#define  __doc__funclet "funclet(code)\n"\
"Download a 'funclet' contained in the string 'code' to the target\n"\
"and execute it. This function waits until the code stops on a \"jmp $\"\n"\
"or a timeout.\n"\
"Please refer to the 'funclet' documentation for the contents of the\n"\
"code string.\n"
static PyObject*
py_funclet(PyObject *self, PyObject *args) {
    STATUS_T status;
    char * code;
    int size;
    if (!PyArg_ParseTuple(args, "s#", &code, &size))
        return 0;
    if (size & 1) {
        PyErr_Format(PyExc_ValueError, "data must be of even size");
        return NULL;
    }
    
    status = MSP430_Funclet(code, size, 1, 1);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not execute code");
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

#define  __doc__configure "configure(mode, value)\n"\
"Configure the MSP430 driver.\n"
static PyObject*
py_configure(PyObject *self, PyObject *args) {
    STATUS_T status;
    int mode;
    int value;
    if (!PyArg_ParseTuple(args, "ii", &mode, &value))
        return 0;
    
    status = MSP430_Configure(mode, value);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not change mode");
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

#define  __doc__regread "regread(regnum)\n"\
"returns register value\n"
static PyObject*
py_regread(PyObject *self, PyObject *args) {
    STATUS_T status;
    LONG value;
    int regnum;
    if (!PyArg_ParseTuple(args, "i", &regnum))
        return 0;
    
    status = MSP430_ReadRegister(regnum, &value);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not read register");
        return NULL;
    }
    return Py_BuildValue("i", value);
}

#define  __doc__regwrite "regwrite(regnum, value)\n"\
"write value to register\n"
static PyObject*
py_regwrite(PyObject *self, PyObject *args) {
    STATUS_T status;
    int value;
    int regnum;
    if (!PyArg_ParseTuple(args, "ii", &regnum, &value))
        return 0;
    
    status = MSP430_WriteRegister(regnum, value);
    if (status != STATUS_OK) {
        PyErr_Format(PyExc_IOError, "Could not write register");
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *flash_callback_fu = NULL;

#define  __doc__flashcb "set_flash_callback(function)\n"\
"The 'function' is called with (count, total) as arguments\n"\
"while the flash is written.\n"
static PyObject *
py_flashcb(PyObject *dummy, PyObject *args)
{
    PyObject *temp;

    if (PyArg_ParseTuple(args, "O:set_flash_callback", &temp)) {
        if (!PyCallable_Check(temp)) {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }
        Py_XINCREF(temp);         /* Add a reference to new callback */
        Py_XDECREF(flash_callback_fu);    /* Dispose of previous callback */
        flash_callback_fu = temp;         /* Remember new callback */
    }
    Py_INCREF(Py_None);
    return Py_None;
}

void messagecallback(WORD count, WORD total) {
    /* Time to call the callback */
    if (flash_callback_fu) {
        PyObject *arglist;
        PyObject *result;
        arglist = Py_BuildValue("(ii)", count, total);
        result = PyEval_CallObject(flash_callback_fu, arglist);
        Py_DECREF(arglist);
        if (result == NULL) {
            fprintf(stderr,"Exception in callback:\n");
            PyErr_Print();
            return;
        }
        /* Here maybe use the result */
        Py_DECREF(result);
    }
}

#define  __doc__ishalted "isHalted()\n"\
"Check if cpu is stuck on an address.\n"
static PyObject*
py_ishalted(PyObject *self, PyObject *args) {
    int value;
    if (!PyArg_ParseTuple(args, ""))
        return 0;
    
    value = MSP430_isHalted();
    return Py_BuildValue("i", value);
}

static PyMethodDef parjtag_methods[] = {
    //JTAG functions exposed to python
    {"connect",         py_connect,     METH_VARARGS, __doc__connect},
    {"release",         py_release,     METH_VARARGS, __doc__release},
    {"reset",           py_reset,       METH_VARARGS, __doc__reset},
    {"memread",         py_memread,     METH_VARARGS, __doc__memread},
    {"memwrite",        py_memwrite,    METH_VARARGS, __doc__memwrite},
    {"memverify",       py_memverify,   METH_VARARGS, __doc__memverify},
    {"memerase",        py_memerase,    METH_VARARGS, __doc__memerase},
    {"funclet",         py_funclet,     METH_VARARGS, __doc__funclet},
    {"configure",       py_configure,   METH_VARARGS, __doc__configure},
    {"regread",         py_regread,     METH_VARARGS, __doc__regread},
    {"regwrite",        py_regwrite,    METH_VARARGS, __doc__regwrite},
    {"set_flash_callback", py_flashcb,  METH_VARARGS, __doc__flashcb},
    {"isHalted",        py_ishalted,    METH_VARARGS, __doc__ishalted},
    //End maeker
    {0, 0}
};

/* module entry-point (module-initialization) function */
void init_parjtag(void) {
    //Create the module and add the functions
    Py_InitModule("_parjtag", parjtag_methods);
    //register callback function
    MSP430_Configure(FLASH_CALLBACK, (LONG)&messagecallback);
}
